package sun.rmi.transport.tcp;

import sun.rmi.runtime.RuntimeUtil;
import sun.security.action.GetIntegerAction;
import sun.security.action.GetLongAction;
import java.util.ListIterator;
import java.util.concurrent.TimeUnit;
import java.net.Socket;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.rmi.ConnectIOException;
import java.rmi.RemoteException;
import sun.rmi.runtime.Log;
import sun.rmi.transport.Connection;
import java.lang.ref.SoftReference;
import java.security.AccessController;
import sun.rmi.transport.Endpoint;
import java.util.ArrayList;
import java.util.concurrent.ScheduledExecutorService;
import java.lang.ref.Reference;
import java.util.WeakHashMap;
import java.security.AccessControlContext;
import java.util.concurrent.Future;
import java.util.List;
import sun.rmi.transport.Channel;

public class TCPChannel implements Channel
{
    private final TCPEndpoint ep;
    private final TCPTransport tr;
    private final List<TCPConnection> freeList;
    private Future<?> reaper;
    private boolean usingMultiplexer;
    private ConnectionMultiplexer multiplexer;
    private ConnectionAcceptor acceptor;
    private AccessControlContext okContext;
    private WeakHashMap<AccessControlContext, Reference<AccessControlContext>> authcache;
    private SecurityManager cacheSecurityManager;
    private static final long idleTimeout;
    private static final int handshakeTimeout;
    private static final int responseTimeout;
    private static final ScheduledExecutorService scheduler;
    
    TCPChannel(TCPTransport tr, TCPEndpoint ep) {
        this.freeList = new ArrayList();
        this.reaper = null;
        this.usingMultiplexer = false;
        this.multiplexer = null;
        this.cacheSecurityManager = null;
        this.tr = tr;
        this.ep = ep;
    }
    
    @Override
    public Endpoint getEndpoint() {
        return this.ep;
    }
    
    private void checkConnectPermission() throws SecurityException {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager == null) {
            return;
        }
        if (securityManager != this.cacheSecurityManager) {
            this.okContext = null;
            this.authcache = new WeakHashMap();
            this.cacheSecurityManager = securityManager;
        }
        AccessControlContext context = AccessController.getContext();
        if (this.okContext == null || (!this.okContext.equals(context) && !this.authcache.containsKey(context))) {
            securityManager.checkConnect(this.ep.getHost(), this.ep.getPort());
            this.authcache.put(context, new SoftReference(context));
        }
        this.okContext = context;
        return;
    }
    
    @Override
    public Connection newConnection() throws RemoteException {
        TCPConnection tcpConnection;
        do {
            tcpConnection = null;
            synchronized (this.freeList) {
                int n = this.freeList.size() - 1;
                if (n >= 0) {
                    this.checkConnectPermission();
                    tcpConnection = (TCPConnection)this.freeList.get(n);
                    this.freeList.remove(n);
                }
            }
            if (tcpConnection != null) {
                if (!tcpConnection.isDead()) {
                    TCPTransport.tcpLog.log(Log.BRIEF, "reuse connection");
                    return tcpConnection;
                }
                this.free(tcpConnection, false);
            }
            else {
                continue;
            }
        } while (tcpConnection != null);
        return this.createConnection();
    }
    
    private Connection createConnection() throws RemoteException {
        TCPTransport.tcpLog.log(Log.BRIEF, "create connection");
        TCPConnection openConnection;
        if (!this.usingMultiplexer) {
            Socket socket = this.ep.newSocket();
            openConnection = new TCPConnection(this, socket);
            try {
                DataOutputStream dataOutputStream = new DataOutputStream(openConnection.getOutputStream());
                this.writeTransportHeader(dataOutputStream);
                if (!openConnection.isReusable()) {
                    dataOutputStream.writeByte(76);
                }
                else {
                    dataOutputStream.writeByte(75);
                    dataOutputStream.flush();
                    int soTimeout = 0;
                    try {
                        soTimeout = socket.getSoTimeout();
                        socket.setSoTimeout(TCPChannel.handshakeTimeout);
                    }
                    catch (Exception ex2) {}
                    DataInputStream o = new DataInputStream(openConnection.getInputStream());
                    int byte1 = o.readByte();
                    if (byte1 != 78) {
                        throw new ConnectIOException((byte1 == 79) ? "JRMP StreamProtocol not supported by server" : "non-JRMP server at remote endpoint");
                    }
                    String utf = o.readUTF();
                    int int1 = o.readInt();
                    if (TCPTransport.tcpLog.isLoggable(Log.VERBOSE)) {
                        TCPTransport.tcpLog.log(Log.VERBOSE, "server suggested " + utf + ":" + int1);
                    }
                    TCPEndpoint.setLocalHost(utf);
                    TCPEndpoint localEndpoint = TCPEndpoint.getLocalEndpoint(0, null, null);
                    dataOutputStream.writeUTF(localEndpoint.getHost());
                    dataOutputStream.writeInt(localEndpoint.getPort());
                    if (TCPTransport.tcpLog.isLoggable(Log.VERBOSE)) {
                        TCPTransport.tcpLog.log(Log.VERBOSE, "using " + localEndpoint.getHost() + ":" + localEndpoint.getPort());
                    }
                    try {
                        socket.setSoTimeout((soTimeout != 0) ? soTimeout : TCPChannel.responseTimeout);
                    }
                    catch (Exception ex3) {}
                    dataOutputStream.flush();
                }
            }
            catch (IOException ex4) {
                try {
                    openConnection.close();
                }
                catch (Exception ex5) {}
                if ((ex4 instanceof RemoteException)) {
                    throw (RemoteException)ex4;
                }
                throw new ConnectIOException("error during JRMP connection establishment", ex4);
            }
        }
        else {
            try {
                openConnection = this.multiplexer.openConnection();
            }
            catch (IOException ex) {
                synchronized (this) {
                    this.usingMultiplexer = false;
                    this.multiplexer = null;
                }
                throw new ConnectIOException("error opening virtual connection over multiplexed connection", ex);
            }
        }
        return openConnection;
    }
    
    @Override
    public void free(Connection connection, boolean b) {
        if (connection == null) {
            return;
        }
        if (b && connection.isReusable()) {
            long currentTimeMillis = System.currentTimeMillis();
            TCPConnection tcpConnection = (TCPConnection)connection;
            TCPTransport.tcpLog.log(Log.BRIEF, "reuse connection");
            synchronized (this.freeList) {
                this.freeList.add(tcpConnection);
                if (this.reaper == null) {
                    TCPTransport.tcpLog.log(Log.BRIEF, "create reaper");
                    this.reaper = TCPChannel.scheduler.scheduleWithFixedDelay(new Runnable() {
                        @Override
                        public void run() {
                            TCPTransport.tcpLog.log(Log.VERBOSE, "wake up");
                            TCPChannel.access$000(TCPChannel.this);
                        }
                    }, TCPChannel.idleTimeout, TCPChannel.idleTimeout, TimeUnit.MILLISECONDS);
                }
            }
            tcpConnection.setLastUseTime(currentTimeMillis);
            tcpConnection.setExpiration(currentTimeMillis + TCPChannel.idleTimeout);
        }
        else {
            TCPTransport.tcpLog.log(Log.BRIEF, "close connection");
            try {
                connection.close();
            }
            catch (IOException ex) {}
        }
        return;
    }
    
    private void writeTransportHeader(DataOutputStream out) throws RemoteException {
        try {
            DataOutputStream dataOutputStream = new DataOutputStream(out);
            dataOutputStream.writeInt(1246907721);
            dataOutputStream.writeShort(2);
        }
        catch (IOException ex) {
            throw new ConnectIOException("error writing JRMP transport header", ex);
        }
    }
    
    synchronized void useMultiplexer(ConnectionMultiplexer multiplexer) {
        this.multiplexer = multiplexer;
        this.usingMultiplexer = true;
    }
    
    void acceptMultiplexConnection(Connection connection) {
        if (this.acceptor == null) {
            (this.acceptor = new ConnectionAcceptor(this.tr)).startNewAcceptor();
        }
        this.acceptor.accept(connection);
    }
    
    public void shedCache() {
        Connection[] array;
        synchronized (this.freeList) {
            array = (Connection[])this.freeList.toArray(new Connection[this.freeList.size()]);
            this.freeList.clear();
        }
        int length = array.length;
        while (--length >= 0) {
            Connection connection = array[length];
            array[length] = null;
            try {
                connection.close();
            }
            catch (IOException ex) {}
        }
    }
    
    private void freeCachedConnections() {
        synchronized (this.freeList) {
            int size = this.freeList.size();
            if (size > 0) {
                long currentTimeMillis = System.currentTimeMillis();
                ListIterator<TCPConnection> listIterator = this.freeList.listIterator(size);
                while (listIterator.hasPrevious()) {
                    TCPConnection tcpConnection = (TCPConnection)listIterator.previous();
                    if (tcpConnection.expired(currentTimeMillis)) {
                        TCPTransport.tcpLog.log(Log.VERBOSE, "connection timeout expired");
                        try {
                            tcpConnection.close();
                        }
                        catch (IOException ex) {}
                        listIterator.remove();
                    }
                    else {
                        continue;
                    }
                }
            }
            if (this.freeList.isEmpty()) {
                this.reaper.cancel(false);
                this.reaper = null;
            }
        }
    }
    
    static void access$000(TCPChannel tcpChannel) {
        tcpChannel.freeCachedConnections();
    }
    
    static {
        idleTimeout = (long)AccessController.doPrivileged(new GetLongAction("sun.rmi.transport.connectionTimeout", 15000L));
        handshakeTimeout = (int)AccessController.doPrivileged(new GetIntegerAction("sun.rmi.transport.tcp.handshakeTimeout", 60000));
        responseTimeout = (int)AccessController.doPrivileged(new GetIntegerAction("sun.rmi.transport.tcp.responseTimeout", 0));
        scheduler = ((RuntimeUtil)AccessController.doPrivileged(new RuntimeUtil.GetInstanceAction())).getScheduler();
    }
}
